import cherrypy
from dogpile.cache import make_region, register_backend
from dogpile.cache.backends.memory import MemoryBackend


def _setupCache(curConfig: dict):
    """
    Setup caching based on the provided configuration.

    Cache backends are forcibly replaced because Girder initially configures
    the regions with the null backends.
    """
    from girder.settings import SettingKey
    from girder.models.setting import Setting

    if Setting().get(SettingKey.CACHE_ENABLED):
        # Replace existing backend. This is necessary
        # because they're initially configured with the null backend.
        # The other values here can be overridden by the CACHE_CONFIG setting object.
        cacheConfig = {
            'cache.global.replace_existing_backend': True,
            'cache.request.replace_existing_backend': True,
            'cache.global.backend': 'dogpile.cache.memory',
            'cache.request.backend': 'cherrypy_request',
        }

        cacheConfig.update(Setting().get(SettingKey.CACHE_CONFIG))

        cache.configure_from_config(cacheConfig, 'cache.global.')
        requestCache.configure_from_config(cacheConfig, 'cache.request.')
    else:
        # Reset caches back to null cache (in the case of server teardown)
        cache.configure(backend='dogpile.cache.null', replace_existing_backend=True)
        requestCache.configure(backend='dogpile.cache.null', replace_existing_backend=True)

    # Although the rateLimitBuffer has no pre-existing backend, this method may be called multiple
    # times in testing (where caches were already configured)
    rateLimitBuffer.configure(backend='dogpile.cache.memory', replace_existing_backend=True)


class CherrypyRequestBackend(MemoryBackend):
    """
    A memory backed cache for individual CherryPy requests.

    This provides a cache backend for dogpile.cache which is designed
    to work in a thread-safe manner using cherrypy.request, a thread local
    storage that only lasts for the duration of a request.
    """

    def __init__(self, arguments):
        pass

    @property
    def _cache(self):
        if not hasattr(cherrypy.request, '_girderCache'):
            cherrypy.request._girderCache = {}

        return cherrypy.request._girderCache


register_backend('cherrypy_request', 'girder.utility._cache', 'CherrypyRequestBackend')

# These caches must be configured with the null backend upon creation due to the fact
# that user-based configuration of the regions doesn't happen until server start, which
# doesn't occur when using Girder as a library.
cache = make_region(name='girder.cache').configure(backend='dogpile.cache.null')
requestCache = make_region(name='girder.request').configure(backend='dogpile.cache.null')

# This cache is not configurable by the user, and will always be configured when the server is.
# It holds data for rate limiting, which is ephemeral, but must be persisted (i.e. it's not optional
# or best-effort).
# TODO we need to make this configurable to be share-nothing in a multi-process environment
rateLimitBuffer = make_region(name='girder.rate_limit').configure(backend='dogpile.cache.memory')



import datetime
import time

from .model_base import Model


class ProgressState:
    """
    Enum of possible progress states for progress records.
    """

    QUEUED = 'queued'
    ACTIVE = 'active'
    SUCCESS = 'success'
    ERROR = 'error'

    @classmethod
    def isComplete(cls, state):
        return state == cls.SUCCESS or state == cls.ERROR


class Notification(Model):
    """
    This model is used to represent a notification that should be streamed
    to a specific user in some way. Each notification contains a
    type field indicating what kind of notification it is, a userId field
    indicating which user the notification should be sent to, a data field
    representing the payload of the notification, a time field indicating the
    time at which the event happened, and an optional expires field indicating
    at what time the notification should be deleted from the database.
    """

    def initialize(self):
        self.name = 'notification'
        self.ensureIndices(('userId', 'time', 'updated', 'tokenId'))
        self.ensureIndex(('expires', {'expireAfterSeconds': 0}))

    def validate(self, doc):
        return doc

    def createNotification(self, type, data, user, expires=None, token=None):
        """
        Create a generic notification.

        :param type: The notification type.
        :type type: str
        :param data: The notification payload.
        :param user: User to send the notification to.
        :type user: dict
        :param expires: Expiration date (for transient notifications).
        :type expires: datetime.datetime
        :param token: Set this if the notification should correspond to a token
            instead of a user.
        :type token: dict
        """
        now = datetime.datetime.now(datetime.timezone.utc)
        currentTime = time.time()
        doc = {
            'type': type,
            'data': data,
            'time': now,
            'updated': now,
            'startTime': currentTime,
            'updatedTime': currentTime
        }
        if user:
            doc['userId'] = user['_id']
        elif token:
            doc['tokenId'] = token['_id']

        if expires is not None:
            doc['expires'] = expires

        return self.save(doc)

    def initProgress(self, user, title, total=0, state=ProgressState.ACTIVE,
                     current=0, message='', token=None, estimateTime=True, resource=None,
                     resourceName=None):
        """
        Create a "progress" type notification that can be updated anytime there
        is progress on some task. Progress records that are not updated for more
        than one hour will be deleted. The "time" field of a progress record
        indicates the time the task was started.

        :param user: the user associated with this notification.  If this is
            None, a session token must be specified.
        :param title: The title of the task. This should not change over the
            course of the task. (e.g. 'Deleting folder "foo"')
        :type title: str
        :param total: Some numeric value representing the total task length. By
            convention, setting this <= 0 means progress on this task is
            indeterminate.
        :type total: int, long, or float
        :param state: Represents the state of the underlying task execution.
        :type state: ProgressState enum value.
        :param current: Some numeric value representing the current progress of
            the task (relative to total).
        :type current: int, long, or float
        :param message: Message corresponding to the current state of the task.
        :type message: str
        :param token: if the user is None, associate this notification with the
            specified session token.
        :param estimateTime: if True, generate an estimate of the total time
            the task will take, if possible.  If False, never generate a time
            estimate.
        :param resource: a partial or complete resource that the notification is
            associated with. This must at a minimum include the id of the resource.
        :param resourceName: the type of resource the notification is associated with.
        """
        data = {
            'title': title,
            'total': total,
            'current': current,
            'state': state,
            'message': message,
            'estimateTime': estimateTime,
            'resource': resource,
            'resourceName': resourceName
        }
        expires = datetime.datetime.now(datetime.timezone.utc) + datetime.timedelta(hours=1)

        return self.createNotification('progress', data, user, expires,
                                       token=token)

    def updateProgress(self, record, save=True, **kwargs):
        """
        Update an existing progress record.

        :param record: The existing progress record to update.
        :type record: dict
        :param total: Some numeric value representing the total task length. By
            convention, setting this <= 0 means progress on this task is
            indeterminate. Generally this shouldn't change except in cases where
            progress on a task switches between indeterminate and determinate
            state.
        :type total: int, long, or float
        :param state: Represents the state of the underlying task execution.
        :type state: ProgressState enum value.
        :param current: Some numeric value representing the current progress
            of the task (relative to total).
        :type current: int, long, or float
        :param increment: Amount to increment the progress by. Don't pass both
            current and increment together, as that behavior is undefined.
        :type increment: int, long, or float
        :param message: Message corresponding to the current state of the task.
        :type message: str
        :param expires: Set a custom (UTC) expiration time on the record.
            Default is one hour from the current time.
        :type expires: datetime
        :param save: Whether to save the record to the database.
        :type save: bool
        """
        if 'increment' in kwargs:
            record['data']['current'] += kwargs['increment']

        for field, value in kwargs.items():
            if field in ('total', 'current', 'state', 'message'):
                record['data'][field] = value

        now = datetime.datetime.now(datetime.timezone.utc)

        if 'expires' in kwargs:
            expires = kwargs['expires']
        else:
            expires = now + datetime.timedelta(hours=1)

        record['updated'] = now
        record['expires'] = expires
        record['updatedTime'] = time.time()
        if save:
            # Only update the time estimate if we are also saving
            if (record['updatedTime'] > record['startTime']
                    and record['data']['estimateTime']):
                if 'estimatedTotalTime' in record:
                    del record['estimatedTotalTime']
                try:
                    total = float(record['data']['total'])
                    current = float(record['data']['current'])
                    if total >= current and total > 0 and current > 0:
                        record['estimatedTotalTime'] = \
                            total * (record['updatedTime'] - record['startTime']) / current
                except ValueError:
                    pass
            return self.save(record)
        else:
            return record

    def get(self, user, since=None, token=None, sort=None):
        """
        Get outstanding notifications for the given user.

        :param user: The user requesting updates.  None to use the token
            instead.
        :param since: Limit results to entities that have been updated
            since a certain timestamp.
        :type since: datetime
        :param token: if the user is None, the token requesting updated.
        :param sort: Sort field for the database query.
        """
        q = {}
        if user:
            q['userId'] = user['_id']
        else:
            q['tokenId'] = token['_id']

        if since is not None:
            q['updated'] = {'$gt': since}

        return self.find(q, sort=sort)



from girder.models.assetstore import Assetstore
from girder.utility.progress import ProgressContext
from girder_worker.app import app


@app.task(queue='local')
def importDataTask(
    assetstore: dict,
    parent: dict,
    parentType: str,
    params: dict,
    progress: bool,
    user: dict,
    leafFoldersAsItems: bool
):
    with ProgressContext(progress, user=user, title='Importing data') as ctx:
        Assetstore().importData(
            assetstore, parent=parent, parentType=parentType, params=params,
            progress=ctx, user=user, leafFoldersAsItems=leafFoldersAsItems,
        )



import logging
import pymongo
import pymongo.cursor
import urllib.parse

from girder.utility import config

_dbClients = {}
logger = logging.getLogger(__name__)


if not hasattr(pymongo.cursor.Cursor, 'count'):
    import warnings

    def _cursorCount(self, with_limit_and_skip=False):
        warnings.warn(
            'count is deprecated. Use Collection.count_documents instead.',
            DeprecationWarning,
            stacklevel=2,
        )
        params = {}
        if with_limit_and_skip and getattr(self, '_limit', getattr(self, '_Cursor__limit', None)):
            params['limit'] = getattr(self, '_limit', getattr(self, '_Cursor__limit', None))
        if with_limit_and_skip and getattr(self, '_skip', getattr(self, '_Cursor__skip', None)):
            params['skip'] = getattr(self, '_skip', getattr(self, '_Cursor__skip', None))
        return getattr(self, '_collection', getattr(
            self, '_Cursor__collection', None)).count_documents(
                getattr(self, '_spec', getattr(self, '_Cursor__spec', None)), **params)

    pymongo.cursor.Cursor.count = _cursorCount


def getDbConfig():
    """Get the database configuration values from the cherrypy config."""
    cfg = config.getConfig()
    if 'database' in cfg:
        return cfg['database']
    else:
        return {}


def getDbConnection(uri=None, replicaSet=None, quiet=False, **kwargs):
    """
    Get a MongoClient object that is connected to the configured database.
    We lazy-instantiate a module-level singleton, the MongoClient objects
    manage their own connection pools internally. Any extra kwargs you pass to
    this method will be passed through to the MongoClient.

    :param uri: if specified, connect to this mongo db rather than the one in
                the config.
    :param replicaSet: if uri is specified, use this replica set.
    :param quiet: if true, don't log warnings and success.
    :type quiet: bool
    """
    global _dbClients

    origKey = (uri, replicaSet)
    if origKey in _dbClients:
        return _dbClients[origKey]

    dbConf = getDbConfig()

    if uri is None or uri == '':
        uri = dbConf.get('uri')
        replicaSet = dbConf.get('replica_set')

    clientOptions = {
        # This is the maximum time between when we fetch data from a cursor.
        # If it times out, the cursor is lost and we can't reconnect.  If it
        # isn't set, we have issues with replica sets when the primary goes
        # down.  This value can be overridden in the mongodb uri connection
        # string with the socketTimeoutMS.
        'socketTimeoutMS': 60000,
        'connectTimeoutMS': 20000,
        'serverSelectionTimeoutMS': 20000,
        'readPreference': 'secondaryPreferred',
        'replicaSet': replicaSet,
        'w': 'majority'
    }

    # All other options in the [database] section will be passed directly as
    # options to the mongo client
    for opt, val in dict(dbConf).items():
        if opt not in {'uri', 'replica_set'}:
            clientOptions[opt] = val

    # Finally, kwargs take precedence
    clientOptions.update(kwargs)
    # if the connection URI overrides any option, honor it above our own
    # settings.
    uriParams = urllib.parse.parse_qs(urllib.parse.urlparse(uri).query)
    for key in uriParams:
        if key in clientOptions:
            del clientOptions[key]

    if uri is None:
        dbUriRedacted = 'mongodb://localhost:27017/girder'
        if not quiet:
            logger.warning('WARNING: No MongoDB URI specified, using the default value')

        client = pymongo.MongoClient(dbUriRedacted, **clientOptions)
    else:
        parts = uri.split('@')
        if len(parts) == 2:
            dbUriRedacted = 'mongodb://' + parts[1]
        else:
            dbUriRedacted = uri

        client = pymongo.MongoClient(uri, **clientOptions)

    if not quiet:
        desc = ''
        if replicaSet:
            desc += ', replica set: %s' % replicaSet
        logger.info('Connecting to MongoDB: %s%s', dbUriRedacted, desc)

    # Make sure we can connect to the mongo server at startup
    client.server_info()

    _dbClients[origKey] = _dbClients[(uri, replicaSet)] = client

    return client



from pathlib import Path
import subprocess
import sys

plugins_dir = Path(sys.argv[1])

for plugin_dir in plugins_dir.iterdir():
    if plugin_dir.is_dir():
        search_dirs = [
            plugin_dir / f'girder_{plugin_dir.name}' / 'web_client',
            plugin_dir / f'girder_plugin_{plugin_dir.name}' / 'web_client'
        ]

        for web_client_dir in search_dirs:
            if web_client_dir.exists() and (web_client_dir / 'package-lock.json').exists():
                subprocess.run(
                    'npm ci && SKIP_SOURCE_MAPS=true npm run build',
                    check=True,
                    shell=True,
                    stdout=sys.stdout,
                    stderr=sys.stderr,
                    cwd=web_client_dir
                )



from click_plugins import with_plugins
from pkg_resources import iter_entry_points
import click


@with_plugins(iter_entry_points('girder.cli_plugins'))
@click.group(help='Girder: data management platform for the web.',
             context_settings=dict(help_option_names=['-h', '--help']))
@click.version_option(message='%(version)s')
def main():
    pass



import logging
import os
from pkg_resources import DistributionNotFound, get_distribution

from .utility import config

try:
    __version__ = get_distribution(__name__).version
except DistributionNotFound:
    # package is not installed
    __version__ = None

__license__ = 'Apache 2.0'
auditLogger = logging.getLogger('girder_audit')
auditLogger.setLevel(logging.INFO)
auditLogger.propagate = False

if not os.getenv('GIRDER_SPHINX_BUILD'):
    config.loadConfig()  # Populate the config info at import time
